/********************************************
 * UTIA AV CR, v.v.i
 * Signal Processing Department
 * Pod Vodarenskou vezi 4
 * 182 08 Prague
 * Czech Republic
 * web: http://zs.utia.cas.cz
 * contact: sp@utia.cz
 *
 * Filename: i2c_im.c
 * Date: 2022.03.03
 * Author: Lukas Kohout <kohoutl@utia.cas.cz>
 * Description:
 *    I2C controller init
 *
 *
 * Changelog:
 *
 *
 *
 *
 ********************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <sys/ioctl.h>

#include "i2c_im.h"
#include "../gpio/gpio.h"


/*
 * I2C write register - 1B register address, 1B register Size
 */
int i2c_im_write(int i2c_fd, unsigned char dev_addr, unsigned char reg, unsigned char data)
{
    int           status;
    unsigned char buff[2];

    struct i2c_msg              msgs[1];
    struct i2c_rdwr_ioctl_data  msgset[1];

    buff[0] = reg;
    buff[1] = data;

    msgs[0].addr  = dev_addr;
    msgs[0].flags = 0;
    msgs[0].len   = 2;
    msgs[0].buf   = buff;

    msgset[0].msgs = msgs;
    msgset[0].nmsgs = 1;

    status = ioctl(i2c_fd, IM_I2C_RDWR, &msgset);
    if (status < 0) {
        return -1;
    }

    return 0;
}


/*
 * I2C read register - 1B register address, 1B register Size
 */
int i2c_im_read(int i2c_fd, unsigned char dev_addr, unsigned char reg)
{
    int           status;
    unsigned char buff_wr[1];
    unsigned char buff_rd[1];

    struct i2c_msg              msgs[2];
    struct i2c_rdwr_ioctl_data  msgset[1];

    buff_wr[0] = reg;

    msgs[0].addr = dev_addr;
    msgs[0].flags = 0;
    msgs[0].len = 1;
    msgs[0].buf = buff_wr;

    msgs[1].addr = dev_addr;
    msgs[1].flags = I2C_M_RD;
    msgs[1].len = 1;
    msgs[1].buf = buff_rd;

    msgset[0].msgs = msgs;
    msgset[0].nmsgs = 2;


    // set slave address
    status = ioctl(i2c_fd, IM_I2C_RDWR, &msgset);
    if (status < 0) {
        return -1;
    }

    return (int)*buff_rd;
}


/*
 * configure I2C multiplexer
 */
int i2c_im_sel(int i2c_fd, unsigned char sel)
{
    int           status;
    unsigned char buff[1];

    struct i2c_msg              msgs[1];
    struct i2c_rdwr_ioctl_data  msgset[1];

    buff[0] = sel;

    msgs[0].addr  = IM_MUX_ADDR;
    msgs[0].flags = 0;
    msgs[0].len   = 1;
    msgs[0].buf   = buff;

    msgset[0].msgs = msgs;
    msgset[0].nmsgs = 1;

    status = ioctl(i2c_fd, IM_I2C_RDWR, &msgset);
    if (status < 0) {
        return -1;
    }

    return 0;
}


/*
 * reset AVNET board
 *  - EMIO GPIO
 */
int i2c_im_rst(im_reset_t *rst)
{

    int status;

    gpio_pin_t gpio_channel_list[1] = {
        {rst->pin, 1}   // FMC resetn
    };

    /* open GPIO */
    status = open_gpio_channels(rst->gpio_base, gpio_channel_list, 1);
    if (status < 0 ) {
        return -1;
    }


    /* Set FMC RESETN */
    status = set_gpio_value(rst->gpio_base, gpio_channel_list, 0, ~IM_RSTN);
    if (status < 0 ) {
        return -2;
    }

    usleep(10000);

    /* Release FMC RESETN */
    status = set_gpio_value(rst->gpio_base, gpio_channel_list, 0, IM_RSTN);
    if (status < 0 ) {
        return -3;
    }


    /* close GPIO */
    status = close_gpio_channels(rst->gpio_base, gpio_channel_list, 1);
    if (status < 0 ) {
        return -4;
    }

    return 0;
}


/*
 * initialize I2C controller, On success it return I2C device FD
 */
int i2c_im_init(char *i2c_device, im_reset_t *rst)
{
    int status;
    int i2c_fd;


    /*
     * reset AVNET board
     */
    status = i2c_im_rst(rst);
    if (status < 0) {
        printf("Cannot reset ... ");
        return -1;
    }


    /*
     * Open I2C device
     */
    i2c_fd = open(i2c_device, O_RDWR);
    if (i2c_fd < 0) {
        printf("Cannot open %s ... ", i2c_device);
        return -2;
    }


    /*
     * Set I2C MUX to IM_MUX_IO_EXP_DATA
     */
    status = i2c_im_sel(i2c_fd, IM_MUX_IO_EXP_DATA);
    if ( status < 0) {
        printf("Cannot select AVNET IO Expander ... ");
        return -3;
    }


    /*
     * configure AVNET IO Expander
     *
     * P7 => unused      = 1 (input)
     * P6 => unused      = 1 (input)
     * P5 => unused      = 1 (input)
     * P4 => HDMIO_PD    = 0 (output)
     * P3 => HDMIO_HPD   = 1 (input)
     * P2 => HDMII_HPD   = 0 (output)
     * P1 => HDMII_INT1  = 1 (input)
     * P0 => HDMII_RST#  = 0 (output)
     *
     */
    status = i2c_im_write(i2c_fd, IM_I2C_IO_EXP_ADDR, 0x03, 0xEA);
    if (status < 0) {
        printf("Cannot configure IO Expander directions ... ");
        return -4;
    }

    //de-assert hdmii reset
    status = i2c_im_write(i2c_fd, IM_I2C_IO_EXP_ADDR, 0x01, 0x01);
    if (status < 0) {
        printf("Cannot configure IO Expander Outs ... ");
        return -5;
    }

    return i2c_fd;
}

/*
 * Close I2C device, release FD
 */
void i2c_im_deinit(int *i2c_fd)
{
    int status;

    do {
        status = close(*i2c_fd);
    } while (status != -1);

    *i2c_fd = -1;
}


