/********************************************
 * UTIA AV CR, v.v.i
 * Signal Processing Department
 * Pod Vodarenskou vezi 4
 * 182 08 Prague
 * Czech Republic
 * web: http://zs.utia.cas.cz
 * contact: sp@utia.cz
 *
 * Filename: vtc.c
 *
 * Date: 2022.03.07
 * Author: Lukas Kohout <kohoutl@utia.cas.cz>
 * Description:
 *    Implements functions for handling Video Timing Controller
 *
 * Changelog:
 *
 *
 *
 *
 ********************************************/

#include <stdio.h>
#include <stdlib.h>
// #include <sleep.h>

#include <string.h>

// #include <xparameters.h>
#include "vtc.h"
#include "xvtc.h"


vtc_timing_t vtc_resolutions[VTC_RES_CNT] = {
    // name              , vav,   vfp,  vsw,   vbp  vsp, vblp,    hav,  hfp,  hsw,    hbp,  hsp,  ,   pclk
    { "1920x1080p60"     , 1080,   3,    5,    37,   1,    1,    1920,   88,   44,    148,   1,    1,    148.50 },
    { "1920x1200p50"     , 1200,   3,    6,    26,   0,    1,    1920,   49,   32,     79,   0,    1,    128.44 },
    { "1600x1200p50"     , 1200,   1,    3,    46,   1,    1,    1600,   64,  192,    304,   1,    1,    135.00 },
    { "1680x1050p60"     , 1050,   3,    6,    21,   0,    1,    1680,   48,   32,     80,   0,    1,    119.23 },
    { "1280x1024p60"     , 1024,   1,    3,    38,   1,    1,    1280,   48,  112,    248,   1,    1,    108.00 },
    { "1280x720p60"      ,  720,   4,    5,    21,   1,    1,    1280,  110,   40,    220,   1,    1,     74.25 },
    { "1024x768p60"      ,  768,   3,    6,    29,   0,    1,    1024,   24,  136,    160,   0,    1,     65.00 },
    { "800x600p60"       ,  600,   1,    4,    23,   1,    1,     800,   40,  128,     88,   1,    1,     40.00 },
    { "640x480p60"       ,  480,   2,    18,   25,   0,    1,     640,    8,   96,     56,   0,    1,     25.19 },  //25.19
    { "600x800p60"       ,  800,   1,    4,    24,   1,    1,     600,   40,  128,     37,   1,    1,     40.00 }
};


/*
 * Function checks lock of vtc detector
 */
uint32_t vtc_is_locked(XVtc *inst)
{
    uint32_t val;

    // clear ERROR register
    val = XVtc_ReadReg(inst->Config.BaseAddress, XVTC_ERROR);
    XVtc_WriteReg(inst->Config.BaseAddress, XVTC_ERROR, val & ~XVTX_ERR_ALL_MASK);

    // clear STATUS register - all interrupts and locks
    val = XVtc_ReadReg(inst->Config.BaseAddress, XVTC_ISR);
    XVtc_WriteReg(inst->Config.BaseAddress, XVTC_ISR, val & ~XVTC_IXR_ALLINTR_MASK);

    // Read Status register again to get LOCK status
    val = XVtc_ReadReg(inst->Config.BaseAddress, XVTC_ISR);
    return (val & XVTC_IXR_LO_MASK);  // bit 8 - LOCK status
}


/*
 * Function read vtc status register
 */
uint32_t vtc_get_status(XVtc *inst)
{
    // Read Status register value back
    return XVtc_ReadReg(inst->Config.BaseAddress, XVTC_ISR);
}


/*
 * Function read vtc error register
 */
uint32_t vtc_get_error(XVtc *inst)
{
  // Read Status register value back
  return XVtc_ReadReg(inst->Config.BaseAddress, XVTC_ERROR);
}


/*
 * VTC SW reset
 */
void vtc_reset(XVtc *inst)
{
    uint32_t val = XVtc_ReadReg(inst->Config.BaseAddress, XVTC_CTL);
    XVtc_WriteReg(inst->Config.BaseAddress, XVTC_CTL, val | XVTC_CTL_RESET_MASK);

    while (XVtc_ReadReg(inst->Config.BaseAddress, XVTC_CTL) & XVTC_CTL_RESET_MASK) {}

}


/*
 *  get timing from VTC detector
 */
void vtc_get_timing(vtc_t *vtc)
{

    // Clear detection structure and setup origin to start of active video
    memset((void *)&vtc->signal, 0, sizeof(XVtc_Signal));
    vtc->signal.OriginMode = 1;
    memset((void *)&vtc->polarity, 0, sizeof(XVtc_Polarity));

    // Setup the VTC Source Select config structure.
    // 1 = Generator registers are source
    // 0 = Detector registers are source
    memset((void *)&vtc->src_sel, 0, sizeof(XVtc_SourceSelect));
    vtc->src_sel.VBlankPolSrc        = 1;
    vtc->src_sel.VSyncPolSrc         = 1;
    vtc->src_sel.HBlankPolSrc        = 1;
    vtc->src_sel.HSyncPolSrc         = 1;
    vtc->src_sel.ActiveVideoPolSrc   = 1;
    vtc->src_sel.ActiveChromaPolSrc  = 1;
    vtc->src_sel.VChromaSrc          = 1;
    vtc->src_sel.VActiveSrc          = 1;
    vtc->src_sel.VBackPorchSrc       = 1;
    vtc->src_sel.VSyncSrc            = 1;
    vtc->src_sel.VFrontPorchSrc      = 1;
    vtc->src_sel.VTotalSrc           = 1;
    vtc->src_sel.HActiveSrc          = 1;
    vtc->src_sel.HBackPorchSrc       = 1;
    vtc->src_sel.HSyncSrc            = 1;
    vtc->src_sel.HFrontPorchSrc      = 1;
    vtc->src_sel.HTotalSrc           = 1;

    // Get Video Input information
    XVtc_GetDetector(&vtc->inst, &vtc->signal);
    XVtc_GetDetectorPolarity(&vtc->inst, &vtc->polarity);

    // Fix the VTC error in detection of the case HFP=0
    if (vtc->signal.HTotal < vtc->signal.HFrontPorchStart) {
        vtc->signal.HFrontPorchStart = vtc->signal.HSyncStart;
    }

    // Fix the VTC detection error for vertical timing
    vtc->signal.V0SyncStart++;
    vtc->signal.V0BackPorchStart++;

}


/*
 * calculate FPS
 */
float vtc_get_fps( vtc_t *vtc, float pclk_mhz )
{
    uint32_t hav = vtc->signal.HFrontPorchStart;
    uint32_t hfp = vtc->signal.HSyncStart - vtc->signal.HFrontPorchStart;
    uint32_t hsw = vtc->signal.HBackPorchStart - vtc->signal.HSyncStart;
    uint32_t hbp = vtc->signal.HTotal - vtc->signal.HBackPorchStart;

    uint32_t vav = vtc->signal.V0FrontPorchStart;
    uint32_t vfp = vtc->signal.V0SyncStart - vtc->signal.V0FrontPorchStart;
    uint32_t vsw = vtc->signal.V0BackPorchStart - vtc->signal.V0SyncStart;
    uint32_t vbp = vtc->signal.V0Total - vtc->signal.V0BackPorchStart;

    uint32_t line_len = hav + hfp + hsw + hbp;
    uint32_t line_cnt = vav + vfp + vsw + vbp;

    float fps = (pclk_mhz * 1000000.00f) / (float)(line_len * line_cnt);

    return fps;
}


/*
 *  preset timing from VTC detector
 */
void vtc_fill_timing(vtc_t *vtc)
{

    // Clear detection structure and setup origin to start of active video
    memset((void *)&vtc->signal, 0, sizeof(XVtc_Signal));
    vtc->signal.OriginMode = 1;
    memset((void *)&vtc->polarity, 0, sizeof(XVtc_Polarity));

    // Setup the VTC Source Select config structure.
    // 1 = Generator registers are source
    // 0 = Detector registers are source
    memset((void *)&vtc->src_sel, 0, sizeof(XVtc_SourceSelect));
    vtc->src_sel.VBlankPolSrc       = 1;
    vtc->src_sel.VSyncPolSrc        = 1;
    vtc->src_sel.HBlankPolSrc       = 1;
    vtc->src_sel.HSyncPolSrc        = 1;
    vtc->src_sel.ActiveVideoPolSrc  = 1;
    vtc->src_sel.ActiveChromaPolSrc = 1;
    vtc->src_sel.VChromaSrc         = 1;
    vtc->src_sel.VActiveSrc         = 1;
    vtc->src_sel.VBackPorchSrc      = 1;
    vtc->src_sel.VSyncSrc           = 1;
    vtc->src_sel.VFrontPorchSrc     = 1;
    vtc->src_sel.VTotalSrc          = 1;
    vtc->src_sel.HActiveSrc         = 1;
    vtc->src_sel.HBackPorchSrc      = 1;
    vtc->src_sel.HSyncSrc           = 1;
    vtc->src_sel.HFrontPorchSrc     = 1;
    vtc->src_sel.HTotalSrc          = 1;

    vtc->polarity.ActiveChromaPol   = 1;
    vtc->polarity.ActiveVideoPol    = 1;


    vtc->polarity.VBlankPol         = vtc_resolutions[vtc->timing_sel].vblp;
    vtc->polarity.VSyncPol          = vtc_resolutions[vtc->timing_sel].vsp;
    vtc->polarity.HBlankPol         = vtc_resolutions[vtc->timing_sel].hblp;
    vtc->polarity.HSyncPol          = vtc_resolutions[vtc->timing_sel].hsp;

    /* Setup the VTC Signal config structure. */
    vtc->signal.HTotal              = vtc_resolutions[vtc->timing_sel].hav +
                                      vtc_resolutions[vtc->timing_sel].hfp +
                                      vtc_resolutions[vtc->timing_sel].hsw +
                                      vtc_resolutions[vtc->timing_sel].hbp;
    vtc->signal.HActiveStart        = 0;
    vtc->signal.HFrontPorchStart    = vtc_resolutions[vtc->timing_sel].hav;  // Active Video Width
    vtc->signal.HSyncStart          = vtc_resolutions[vtc->timing_sel].hav + // Active Video Width + FP Width
                                      vtc_resolutions[vtc->timing_sel].hfp;
    vtc->signal.HBackPorchStart     = vtc_resolutions[vtc->timing_sel].hav + // Active Video Width + FP Width + Sync Width
                                      vtc_resolutions[vtc->timing_sel].hfp +
                                      vtc_resolutions[vtc->timing_sel].hsw;

    vtc->signal.V0Total             = vtc_resolutions[vtc->timing_sel].vav +
                                      vtc_resolutions[vtc->timing_sel].vfp +
                                      vtc_resolutions[vtc->timing_sel].vsw +
                                      vtc_resolutions[vtc->timing_sel].vbp;
    vtc->signal.V0ChromaStart       = 0;
    vtc->signal.V0ActiveStart       = 0;
    vtc->signal.V0FrontPorchStart   = vtc_resolutions[vtc->timing_sel].vav;  // Active Video Height
    vtc->signal.V0SyncStart         = vtc_resolutions[vtc->timing_sel].vav + // Active Video Height + FP Width
                                      vtc_resolutions[vtc->timing_sel].vfp;
    vtc->signal.V0BackPorchStart    = vtc_resolutions[vtc->timing_sel].vav + // Active Video Height + FP Width + Sync Width
                                      vtc_resolutions[vtc->timing_sel].vfp +
                                      vtc_resolutions[vtc->timing_sel].vsw;

}


/*
 * Init vtc generator
 */
void vtc_set_timing(vtc_t *vtc)
{
    XVtc_RegUpdate(&vtc->inst);
    XVtc_SetPolarity(&vtc->inst, &vtc->polarity);
    XVtc_SetGenerator(&vtc->inst, &vtc->signal);
    XVtc_SetSource(&vtc->inst, &vtc->src_sel);
    XVtc_EnableGenerator(&vtc->inst);

}


/*
 * print detected timing from VTC
 */
void vtc_print_timing(vtc_t *vtc)
{
    printf( "VTC Video Signal Information\n");

    printf( "\tHSYNC Timing     = hav=%04d, hfp=%02d, hsw=%02d(hsp=%d), hbp=%03d\n",
              (uint32_t)vtc->signal.HFrontPorchStart,
              (uint32_t)vtc->signal.HSyncStart - vtc->signal.HFrontPorchStart,
              (uint32_t)vtc->signal.HBackPorchStart - vtc->signal.HSyncStart,
              (uint32_t)vtc->polarity.HSyncPol,
              (uint32_t)vtc->signal.HTotal - vtc->signal.HBackPorchStart
    );
    printf( "\tVSYNC Timing     = vav=%04d, vfp=%02d, vsw=%02d(vsp=%d), vbp=%03d\n",
              (uint32_t)vtc->signal.V0FrontPorchStart,
              (uint32_t)vtc->signal.V0SyncStart - vtc->signal.V0FrontPorchStart,
              (uint32_t)vtc->signal.V0BackPorchStart - vtc->signal.V0SyncStart,
              (uint32_t)vtc->polarity.VSyncPol,
              (uint32_t)vtc->signal.V0Total - vtc->signal.V0BackPorchStart
    );
    printf( "\tVideo Dimensions = %d x %d\n",
            (uint32_t)vtc->signal.HFrontPorchStart,
            (uint32_t)vtc->signal.V0FrontPorchStart
    );
}


/*
 * initialize VTC as generator
 */
int vtc_init_common(vtc_t *vtc)
{

    int status;
    XVtc_Config config;// = &(vtc->inst.Config);

    config.BaseAddress = vtc->base_addr;

    // get basic configuration from the xparameters
//     config = XVtc_LookupConfig( dev_id );
//     if ( config == NULL ) {
//         printf("VTC instance with ID %d not fount ... ", dev_id);
//         return -1;
//     }

    // initialize VTC
    status = XVtc_CfgInitialize(&(vtc->inst), &config, config.BaseAddress);
    if (status != 0) {
        printf( "Initialization failed %d ... ", status);
        return -1;
    }
//     printf( "  0x%08X  ", vtc->inst.Config.BaseAddress);

    return 0;
}


void vtc_init_gen(vtc_t *vtc)
{
    //XVtc_Disable( &(vtc->inst) );
    XVtc_DisableGenerator( &(vtc->inst) );
    vtc_fill_timing(vtc);
    vtc_set_timing(vtc);
//     vtc_print_timing(vtc);

}


