/********************************************
 * UTIA AV CR, v.v.i
 * Signal Processing Department
 * Pod Vodarenskou vezi 4
 * 182 08 Prague
 * Czech Republic
 * web: http://zs.utia.cas.cz
 * contact: sp@utia.cz
 *
 * Filename: vdma.c
 * Date: 19.09.2016
 * Author: Lukas Kohout <kohoutl@utia.cas.cz>
 * Description:
 *   SW interface for Xilinx VDMA
 *
 * Changelog:
 *
 *
 *
 *
 ********************************************/

#include <stdio.h>
#include <stdlib.h>

// #include "xparameters.h"

#include "xaxivdma.h"
#include "xaxivdma_hw.h"

#include "vdma.h"

// #define ALIGN_FB


/*
 * VDMA common initialization
 */
int vdma_init_common( vdma_t *vdma, XAxiVdma_Config *vdma_config )
{
    int status;
    XAxiVdma_Config *config = vdma_config;

    // get basic configuration from the xparameters
    config->BaseAddress = vdma->core_base;

//     config = XAxiVdma_LookupConfig( dev_id );
//     if ( !config ) {
//         printf( "VDMA instance with ID %d not found ... ", dev_id);
//         return -1;
//     }

    // Initialize DMA
//     status = XAxiVdma_CfgInitialize(vdma_inst, config, config->BaseAddress);
    status = XAxiVdma_CfgInitialize(&vdma->inst, config, config->BaseAddress);
    if (status != 0) {
        printf( "Initialization failed %d ... ", status);
        return -1;
    }

    //printf("0x%08X ... ", config->BaseAddress);

    if (config->HasMm2S) XAxiVdma_ChannelReset( &(vdma->inst.ReadChannel) );
    if (config->HasS2Mm) XAxiVdma_ChannelReset( &(vdma->inst.WriteChannel) );

//     printf("0x%08X\n", config->BaseAddress);
//     printf("0x%08X\n", vdma->inst.BaseAddr);
//     printf("0x%08X\n", vdma->core_base);
//     printf("%d\n", vdma->inst.AddrWidth);

    return 0;
}


/*
 *  Start VDMA receiver = start writing to the FB
 */
int vdma_rx_start(XAxiVdma *vdma_inst)
{
    int status;

    status = XAxiVdma_DmaStart(vdma_inst, XAXIVDMA_WRITE);
    if (status != 0) {
        //printf( "Cannot start VDMA RX transfer\n", status);
        return -1;
    }

    return 0;
}


/*
 *  Start VDMA transmitter side = start reading from the FB
 */
int vdma_tx_start(XAxiVdma *vdma_inst)
{
  int status;

  status = XAxiVdma_DmaStart(vdma_inst, XAXIVDMA_READ);
  if (status != 0) {
      //printf( "Cannot start VDMA TX transfer\n", status);
      return -1;
  }

  return 0;
}


/*
 *  Stop VDMA receiver = stop writing to the FB
 */
int vdma_rx_stop(XAxiVdma *vdma_inst)
{
    XAxiVdma_DmaStop(vdma_inst, XAXIVDMA_WRITE);
    return 0;
}


/*
 *  Stop VDMA transmitter = stop reading from the FB
 */
int vdma_tx_stop(XAxiVdma *vdma_inst)
{
    XAxiVdma_DmaStop(vdma_inst, XAXIVDMA_READ);
    return 0;
}


/*
 * align line stride to kB
 */
uint32_t vdma_align_stride(uint32_t line_width, u8 byte_cnt)
{
  uint32_t line_bytes = line_width * (uint32_t)byte_cnt;
  uint32_t aligned_line_bytes = 1 << MIN_LINE_BYTES_WIDTH;

  while (aligned_line_bytes < line_bytes) {
      aligned_line_bytes <<= 1;
  }

  return aligned_line_bytes;
}


/*
 * align storage size to kB
 */
uint32_t vdma_align_storage_size(uint32_t line_cnt, uint32_t stride)
{
  uint32_t storage_bytes = line_cnt * stride;
  uint32_t aligned_storage_bytes = 1 << MIN_STORAGE_BYTES_WIDTH;

  while (aligned_storage_bytes < storage_bytes) {
      aligned_storage_bytes <<= 1;
  }

  return aligned_storage_bytes;
}


/*
 * initialize VDMA receiver
 */
int vdma_rx_init( vdma_t *vdma )
{
    int       status;

    int       i;
    uintptr_t addr;
    uint32_t  reg;

    uint32_t  line_len;
    uint32_t  line_cnt;

    uint32_t  storage_stride;
    uint32_t  storage_size;
    uint32_t  storage_offset;

    if ( (vdma->rx_timing.res.x * vdma->rx_timing.res.y) > (vdma->storage_res.x * vdma->storage_res.y) ) {
        printf("\n\tERROR: Frame store is not big enaugh!\n");
        printf("\tFrame store %dx%d (%d B) < Input %dx%d (%d B)\n",
                   vdma->storage_res.x, vdma->storage_res.y, vdma->storage_res.x * vdma->storage_res.y * vdma->rx_pix_bytes,
                   vdma->rx_timing.res.x, vdma->rx_timing.res.y, vdma->rx_timing.res.x * vdma->rx_timing.res.y * vdma->rx_pix_bytes );
        return -1;
    }

    // get active dideo dimensions
    line_cnt        = vdma->rx_timing.res.y;                       // in lines
    line_len        = vdma->rx_timing.res.x * vdma->rx_pix_bytes;  // in bytes

    // get storage dimensions
#ifdef ALIGN_FB
    storage_stride  = vdma_align_stride(vdma->storage_res.x, vdma->rx_pix_bytes);
    storage_size    = vdma_align_storage_size(vdma->storage_res.y, storage_stride);
#else
    storage_stride  = vdma->storage_res.x * vdma->rx_pix_bytes;
    storage_size    = storage_stride * vdma->storage_res.y;
#endif
    storage_offset  = (vdma->rx_offset.x * vdma->rx_pix_bytes) + (storage_stride * vdma->rx_offset.y);

    vdma->rx_setup.VertSizeInput        = line_cnt;
    vdma->rx_setup.HoriSizeInput        = line_len;
    vdma->rx_setup.Stride               = storage_stride;
    vdma->rx_setup.FrameDelay           = 0;    /* This example does not test frame delay */
    vdma->rx_setup.EnableCircularBuf    = 1;
    vdma->rx_setup.EnableSync           = 1;
    vdma->rx_setup.PointNum             = 0;    //1;
    vdma->rx_setup.EnableFrameCounter   = 0;    /* Endless transfers */
    vdma->rx_setup.FixedFrameStoreAddr  = 0;    /* We are not doing parking */

    status = XAxiVdma_DmaConfig(&(vdma->inst), XAXIVDMA_WRITE, &(vdma->rx_setup));
    if (status != 0) {
        printf("Cannot configure VDMA RX channel ... ");
        return -1;
    }

    // Initialize FB addresses
    addr = vdma->fb_base + storage_offset;
    for(i = 0; i < vdma->fb_num; i++) {
        vdma->rx_setup.FrameStoreStartAddr[i] = addr;
        addr += storage_size;
    }

    // Set the buffer addresses for transfer in the DMA engine
    status = XAxiVdma_DmaSetBufferAddr( &(vdma->inst), XAXIVDMA_WRITE, vdma->rx_setup.FrameStoreStartAddr);
    if (status != 0) {
        printf("Cannot set VDMA RX FB addresses ... ");
        return -1;
    }

    // Start the VDMA receiver
    status = vdma_rx_start( &(vdma->inst) );
    if (status != 0) {
        printf("Cannot start VDMA RX transfer ... ");
        return -1;
    }


    // set sync on SOF (TUSER) in AXIS
    status = XAxiVdma_FsyncSrcSelect( &(vdma->inst), XAXIVDMA_S2MM_TUSER_FSYNC, XAXIVDMA_WRITE);
    if (status != 0) {
        printf("Cannot set VDMA RX fsync ... ");
        return -1;
    }


    // enable internal genlock
    addr = vdma->inst.BaseAddr;
    reg = *((volatile int *)(addr+XAXIVDMA_RX_OFFSET+XAXIVDMA_CR_OFFSET ));
    reg |= XAXIVDMA_CR_GENLCK_SRC_MASK;
    *((volatile int *)(addr+XAXIVDMA_RX_OFFSET+XAXIVDMA_CR_OFFSET )) = reg;

    return 0;
}



/*
 * initialize VDMA transmitter
 */
int vdma_tx_init( vdma_t *vdma )
{
    int       status;

    int       i;
    uintptr_t addr;
    uint32_t  reg;

    uint32_t  line_len;
    uint32_t  line_cnt;

    uint32_t  storage_stride;
    uint32_t  storage_size;
    uint32_t  storage_offset;

    if ( (vdma->tx_timing.res.x * vdma->tx_timing.res.y) > (vdma->storage_res.x * vdma->storage_res.y) ) {
      printf("\n\tERROR: Frame store is not big enaugh!\n");
      printf("\tFrame store %dx%d (%d B) < Output %dx%d (%d B)\n",
                vdma->storage_res.x, vdma->storage_res.y, vdma->storage_res.x * vdma->storage_res.y * vdma->tx_pix_bytes,
                vdma->tx_timing.res.x, vdma->tx_timing.res.y, vdma->tx_timing.res.x * vdma->tx_timing.res.y * vdma->tx_pix_bytes );
      return -1;
    }

    // get active dideo dimensions
    line_cnt        = vdma->tx_timing.res.y;                       // in lines
    line_len        = vdma->tx_timing.res.x * vdma->tx_pix_bytes;  // in bytes

    // get storage dimensions
#ifdef ALIGN_FB
    storage_stride  = vdma_align_stride(vdma->storage_res.x, vdma->tx_pix_bytes);
    storage_size    = vdma_align_storage_size(vdma->storage_res.y, storage_stride);
#else
    storage_stride  = vdma->storage_res.x * vdma->tx_pix_bytes;
    storage_size    = storage_stride * vdma->storage_res.y;
#endif
    storage_offset  = (vdma->tx_offset.x * vdma->tx_pix_bytes) + (storage_stride * vdma->tx_offset.y);

    vdma->tx_setup.VertSizeInput        = line_cnt;
    vdma->tx_setup.HoriSizeInput        = line_len;
    vdma->tx_setup.Stride               = storage_stride;
    vdma->tx_setup.FrameDelay           = 0;    /* This example does not test frame delay */
    vdma->tx_setup.EnableCircularBuf    = 1;
    vdma->tx_setup.EnableSync           = 1;
    vdma->tx_setup.PointNum             = 0;    //1;
    vdma->tx_setup.EnableFrameCounter   = 0;    /* Endless transfers */
    vdma->tx_setup.FixedFrameStoreAddr  = 0;    /* We are not doing parking */

    status = XAxiVdma_DmaConfig(&(vdma->inst), XAXIVDMA_READ, &(vdma->tx_setup));
    if (status != 0) {
        printf("Cannot configure VDMA TX channel ... ");
        return -1;
    }

    // Initialize FB addresses
    addr = vdma->fb_base + storage_offset;
    for(i = 0; i < vdma->fb_num; i++) {
        //printf ("%08X  ", addr);
        vdma->tx_setup.FrameStoreStartAddr[i] = addr;
        addr += storage_size;
    }

    // Set the buffer addresses for transfer in the DMA engine
    status = XAxiVdma_DmaSetBufferAddr( &(vdma->inst), XAXIVDMA_READ, vdma->tx_setup.FrameStoreStartAddr);
    if (status != 0) {
        printf("Cannot set VDMA TX FB addresses ... ");
        return -1;
    }

    // Start the VDMA transmitter
    status = vdma_tx_start( &(vdma->inst) );
    if (status != 0) {
        printf("Cannot start VDMA TX transfer ... ");
        return -1;
    }

    // enable internal genlock
    addr = vdma->inst.BaseAddr;
    reg = *((volatile int *)(addr+XAXIVDMA_TX_OFFSET+XAXIVDMA_CR_OFFSET ));
    reg |= XAXIVDMA_CR_GENLCK_SRC_MASK;
    *((volatile int *)(addr+XAXIVDMA_TX_OFFSET+XAXIVDMA_CR_OFFSET )) = reg;

    return 0;
}


/*
 * Clear VDMA error flags
 */
void vdma_clr_errors( vdma_t *vdma )
{
    uintptr_t addr = vdma->inst.BaseAddr;

    // Clear error flags
    *((volatile int *)(addr+XAXIVDMA_RX_OFFSET+XAXIVDMA_SR_OFFSET )) = 0x0000CFF0;
    *((volatile int *)(addr+XAXIVDMA_TX_OFFSET+XAXIVDMA_SR_OFFSET )) = 0x000046F0;
}


/*
 * check VDMA for errors
 */
int vdma_check_err( vdma_t *vdma )
{
    uintptr_t addr = vdma->inst.BaseAddr;

    uint32_t err_rx;
    uint32_t err_tx;
    uint32_t err;

    // Get status of Error Flags
    err_rx = *((volatile int *)(addr+XAXIVDMA_RX_OFFSET+XAXIVDMA_SR_OFFSET )) & 0x0000CFF0;
    err_tx = *((volatile int *)(addr+XAXIVDMA_TX_OFFSET+XAXIVDMA_SR_OFFSET )) & 0x000046F0;
    printf( "VDMA - Checking Error Flags\n" );

    err = (err_rx << 16) | (err_tx);

    if ( err )
    {
        if ( err_rx & 0x00004000 ) printf( "\tS2MM_DMASR - ErrIrq\n"      );
        if ( err_rx & 0x00008000 ) printf( "\tS2MM_DMASR - EOLLateErr\n"  );
        if ( err_rx & 0x00000800 ) printf( "\tS2MM_DMASR - SOFLateErr\n"  );
        if ( err_rx & 0x00000400 ) printf( "\tS2MM_DMASR - SGDecErr\n"    );
        if ( err_rx & 0x00000200 ) printf( "\tS2MM_DMASR - SGSlvErr\n"    );
        if ( err_rx & 0x00000100 ) printf( "\tS2MM_DMASR - EOLEarlyErr\n" );
        if ( err_rx & 0x00000080 ) printf( "\tS2MM_DMASR - SOFEarlyErr\n" );
        if ( err_rx & 0x00000040 ) printf( "\tS2MM_DMASR - DMADecErr\n"   );
        if ( err_rx & 0x00000020 ) printf( "\tS2MM_DMASR - DMASlvErr\n"   );
        if ( err_rx & 0x00000010 ) printf( "\tS2MM_DMASR - DMAIntErr\n"   );

        if ( err_tx & 0x00004000 ) printf( "\tMM2S_DMASR - ErrIrq\n"      );
        if ( err_tx & 0x00000400 ) printf( "\tMM2S_DMASR - SGDecErr\n"    );
        if ( err_tx & 0x00000200 ) printf( "\tMM2S_DMASR - SGSlvErr\n"    );
        if ( err_tx & 0x00000080 ) printf( "\tMM2S_DMASR - SOFEarlyErr\n" );
        if ( err_tx & 0x00000040 ) printf( "\tMM2S_DMASR - DMADecErr\n"   );
        if ( err_tx & 0x00000020 ) printf( "\tMM2S_DMASR - DMASlvErr\n"   );
        if ( err_tx & 0x00000010 ) printf( "\tMM2S_DMASR - DMAIntErr\n"   );

        // Clear error flags
        printf( "AXI_VDMA - Clearing Error Flags\n" );
        vdma_clr_errors(vdma);
    }

    return err;
}


/*
 * print VDMA registers
 */
void vdma_print_registers(vdma_t *vdma)
{
    uintptr_t addr = vdma->inst.BaseAddr;

    // Partial Register Dump
    printf( "VDMA - Registers (addr = %p):\n", (void *)addr );
    printf( "\t PARKPTR          = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_PARKPTR_OFFSET )) );
    printf( "\t ----------------\n" );
    printf( "\t S2MM_DMACR       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_RX_OFFSET+XAXIVDMA_CR_OFFSET )) );
    printf( "\t S2MM_DMASR       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_RX_OFFSET+XAXIVDMA_SR_OFFSET )) );
    printf( "\t S2MM_STRD_FRMDLY = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_S2MM_ADDR_OFFSET+XAXIVDMA_STRD_FRMDLY_OFFSET)) );
    printf( "\t S2MM_START_ADDR0 = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_S2MM_ADDR_OFFSET+XAXIVDMA_START_ADDR_OFFSET+0)) );
    printf( "\t S2MM_START_ADDR1 = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_S2MM_ADDR_OFFSET+XAXIVDMA_START_ADDR_OFFSET+4)) );
    printf( "\t S2MM_START_ADDR2 = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_S2MM_ADDR_OFFSET+XAXIVDMA_START_ADDR_OFFSET+8)) );
    printf( "\t S2MM_HSIZE       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_S2MM_ADDR_OFFSET+XAXIVDMA_HSIZE_OFFSET)) );
    printf( "\t S2MM_VSIZE       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_S2MM_ADDR_OFFSET+XAXIVDMA_VSIZE_OFFSET)) );
    printf( "\t ----------------\n" );
    printf( "\t MM2S_DMACR       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_TX_OFFSET+XAXIVDMA_CR_OFFSET )) );
    printf( "\t MM2S_DMASR       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_TX_OFFSET+XAXIVDMA_SR_OFFSET )) );
    printf( "\t MM2S_STRD_FRMDLY = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_MM2S_ADDR_OFFSET+XAXIVDMA_STRD_FRMDLY_OFFSET)) );
    printf( "\t MM2S_START_ADDR0 = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_MM2S_ADDR_OFFSET+XAXIVDMA_START_ADDR_OFFSET+0)) );
    printf( "\t MM2S_START_ADDR1 = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_MM2S_ADDR_OFFSET+XAXIVDMA_START_ADDR_OFFSET+4)) );
    printf( "\t MM2S_START_ADDR2 = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_MM2S_ADDR_OFFSET+XAXIVDMA_START_ADDR_OFFSET+8)) );
    printf( "\t MM2S_HSIZE       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_MM2S_ADDR_OFFSET+XAXIVDMA_HSIZE_OFFSET)) );
    printf( "\t MM2S_VSIZE       = 0x%08X\n", *((volatile int *)(addr+XAXIVDMA_MM2S_ADDR_OFFSET+XAXIVDMA_VSIZE_OFFSET)) );
    printf( "\t ----------------\n" );
    printf( "\t S2MM_HSIZE_STATUS= 0x%08X\n", *((volatile int *)(addr+0xF0 )) );
    printf( "\t S2MM_VSIZE_STATUS= 0x%08X\n", *((volatile int *)(addr+0xF4 )) );
    printf( "\t ----------------\n" );
}


