// ==============================================================
// Vitis HLS - High-Level Synthesis from C, C++ and OpenCL v2021.2 (64-bit)
// Copyright 1986-2021 Xilinx, Inc. All Rights Reserved.
// ==============================================================
// #ifdef __linux__

/***************************** Include Files *********************************/
#include "vdma.h"

/***************** Macros (Inline Functions) Definitions *********************/
#define MAX_UIO_PATH_SIZE       256
#define MAX_UIO_NAME_SIZE       64
#define MAX_UIO_MAPS            5
#define UIO_INVALID_ADDR        0

/**************************** Type Definitions ******************************/
typedef struct {
    uint64_t addr;
    uint32_t size;
} vdma_uio_map;

typedef struct {
    int           uio_fd;
    int           uio_num;
    char          name[ MAX_UIO_NAME_SIZE ];
    char          version[ MAX_UIO_NAME_SIZE ];
    vdma_uio_map  maps[ MAX_UIO_MAPS ];
} vdma_uio_info;

/***************** Variable Definitions **************************************/
static vdma_uio_info uio_info;


/************************** Function Implementation *************************/
static int line_from_file(char* filename, char* linebuf)
{
    char  *s;
    int   i;
    FILE  *fr;

    fr = fopen(filename, "r");
    if ( !fr ) return -1;

    s = fgets(linebuf, MAX_UIO_NAME_SIZE, fr);
    fclose(fr);

    if ( !s ) return -2;

    for (i = 0; (*s)&&(i<MAX_UIO_NAME_SIZE); i++) {
        if (*s == '\n') *s = 0;
        s++;
    }
    return 0;
}


static int uio_info_read_name(vdma_uio_info* info)
{
    char file[ MAX_UIO_PATH_SIZE ];

    sprintf(file, "/sys/class/uio/uio%d/name", info->uio_num);
    return line_from_file(file, info->name);
}


static int uio_info_read_version(vdma_uio_info* info)
{
    char file[ MAX_UIO_PATH_SIZE ];

    sprintf(file, "/sys/class/uio/uio%d/version", info->uio_num);
    return line_from_file(file, info->version);
}


static int uio_info_read_map_addr(vdma_uio_info* info, int n)
{
    int   status;
    char  file[ MAX_UIO_PATH_SIZE ];
    FILE  *fr;

    info->maps[n].addr = UIO_INVALID_ADDR;
    sprintf(file, "/sys/class/uio/uio%d/maps/map%d/addr", info->uio_num, n);

    fr = fopen(file, "r");
    if (!fr) return -1;

    status = fscanf(fr, "0x%llx", &info->maps[n].addr);
    fclose(fr);

    if (status < 0) return -2;
    return 0;
}


static int uio_info_read_map_size(vdma_uio_info* info, int n)
{
    int   status;
    char  file[ MAX_UIO_PATH_SIZE ];
    FILE  *fr;

    sprintf(file, "/sys/class/uio/uio%d/maps/map%d/size", info->uio_num, n);

    fr = fopen(file, "r");
    if (!fr) return -1;

    status = fscanf(fr, "0x%x", &info->maps[n].size);
    fclose(fr);

    if (status < 0) return -2;
    return 0;
}


int vdma_init(vdma_t *vdma, const char* dev_name)
{
    vdma_uio_info *InfoPtr = &uio_info;
    struct dirent **namelist;
    int i, n;
    char* s;
    char file[ MAX_UIO_PATH_SIZE ];
    char name[ MAX_UIO_NAME_SIZE ];
    int flag = 0;

    assert(vdma != NULL);

    n = scandir("/sys/class/uio", &namelist, 0, alphasort);
    if (n < 0)  return -1;  //XST_DEVICE_NOT_FOUND;

    for (i = 0;  i < n; i++) {
        strcpy(file, "/sys/class/uio/");
        strcat(file, namelist[i]->d_name);
        strcat(file, "/name");
        if ((line_from_file(file, name) == 0) && (strcmp(name, dev_name) == 0)) {
            flag = 1;
            s = namelist[i]->d_name;
            s += 3; // "uio"
            InfoPtr->uio_num = atoi(s);
            break;
        }
    }
    if (flag == 0)  return -1;  //XST_DEVICE_NOT_FOUND;

    uio_info_read_name(InfoPtr);
    uio_info_read_version(InfoPtr);
    for (n = 0; n < MAX_UIO_MAPS; ++n) {
        uio_info_read_map_addr(InfoPtr, n);
        uio_info_read_map_size(InfoPtr, n);
    }

    sprintf(file, "/dev/uio%d", InfoPtr->uio_num);
    printf("/dev/uio%d ... ", InfoPtr->uio_num);
    if ((InfoPtr->uio_fd = open(file, O_RDWR)) < 0) {
        return -2;  //XST_OPEN_DEVICE_FAILED;
    }

    // NOTE: slave interface 'Control' should be mapped to uioX/map0
    vdma->core_base = (uintptr_t)mmap(NULL, InfoPtr->maps[0].size, PROT_READ|PROT_WRITE, MAP_SHARED, InfoPtr->uio_fd, 0 * getpagesize());
    vdma->inst.BaseAddr = vdma->core_base;

    assert(vdma->core_base);

//     vdma->IsReady = XIL_COMPONENT_IS_READY;

    return 0;
}


int vdma_release(vdma_t *vdma)
{
    vdma_uio_info *InfoPtr = &uio_info;

    assert(vdma != NULL);

//     munmap((void*)vdma->base_addr, InfoPtr->maps[0].size);
    munmap((void*)vdma->core_base, InfoPtr->maps[0].size);

    close(InfoPtr->uio_fd);

    return 0;
}

// #endif
